import requests

def mensaje_tel(mensaje: str = 'Termino de correr el script'):
    '''
    Manda un mensaje de telegram al chat que tenes con el bot 'python_scripts'.

    INPUT: 
    mensaje: str: mensaje para mandar al bot.
    '''
    api_token = '5448153732:AAGhKraJQquEqMfpD3cb4rnTcrKB6U1ViMA'
    api_url = f'https://api.telegram.org/bot{api_token}/sendMessage'
    chat_id = '1034347542'
    try:
        response = requests.post(api_url, json={'chat_id': chat_id, 'text': mensaje})
        print(response.text)
    except Exception as e:
        print(e)

if __name__ == '__main__':
    # Se saca de BotFather
    # token = '5448153732:AAGhKraJQquEqMfpD3cb4rnTcrKB6U1ViMA'
    
    # How to get the chat_id
    # url = f'https://api.telegram.org/bot{token}/getUpdates'
    # print(requests.get(url).json())
    # chat_id = 1034347542
    
    # Prueba de si los mensajes funcionan
    mensaje_tel('Los mensajes funcionan')


from sklearn.preprocessing import PolynomialFeatures
from scipy.optimize import curve_fit
from labos.propagacion import Propagacion_errores
import numpy as np, matplotlib.pyplot as plt, re, os

class Ajuste:
    modelos = ['regresion_lineal', 'curve_fit']
    estilos_graficos = ['ajuste_1', 'ajuste_2', 'slider', 'errores'] #TODO: slider no implementado

    def __init__(self, x: np.array, y: np.array, cov_y: np.array = None) -> None:
        '''
        INPUT: 
        x, y: np.arrays: son los datos para ajustar. Si hay más de una variable independiente éstas
        deben disponerse en un solo array cuya dimensión deberá ser X.shape = (numero_de_datos, nume
        ro_de_variables).

        cov_y: np.array: matriz de covarianza de los datos y. También es posible pasar un array con
        las desviaciones estándar de cada dato. Si este es el caso, la matriz de covarianza se const
        ruira automáticamente. De no haber errores, por defecto se toma la identidad.
        '''
        # Chequea que se pasen arrays
        if not(isinstance(x, np.ndarray) and isinstance(y, np.ndarray)):
            raise Exception('Tanto x como y deben ser arrays.')
        
        # Transforma arrays flattens en vectores columna
        if x.ndim == 1:
            self.x = x.reshape(-1,1)
        else:
            self.x = x
        if y.ndim == 1:
            self.y = y.reshape(-1,1)
        else:
            self.y = y
        
        # Determina la matriz de covarianza y el sigma (desviaciones estándar)       
        if cov_y is None:
            # si no se pasa nada, por defecto queda la identidad
            self.cov_y = np.identity(len(self.y))
            self.sigma_y = np.sqrt(np.diag(self.cov_y))

        elif isinstance(cov_y, np.ndarray) and cov_y.ndim != 2:
            raise ValueError("El array que se paso no tiene las dimensiones requeridas.\n Debería tener dimensiones (len(y), len(y)) ó (len(y), 1) si se pasan desviaciones estándar.")
        
        elif isinstance(cov_y, np.ndarray) and cov_y.shape[1] == 1:
            self.cov_y = np.diag(cov_y.reshape(-1)**2)
            self.sigma_y = cov_y
        else:
            self.cov_y = cov_y
            self.sigma_y = np.sqrt(np.diag(cov_y))

        # Hiperparámetros que se definen dentro de la clase
        self.y_modelo = None
        self.vander = None
        self.parametros = None
        self.cov_parametros = None
        self.r = None
        self.R2 = None
        self.chi_2 = None
        self.expr = None
    
    # TODO: implementar distinta forma de llamar 
    # @classmethod # Empleado.foramto_string('Juan-Castro-30000') crea la instancia en ese formato
    # def formato_string(cls, emp_str):
    #   nombre, apellido, sueldo = emp_str.split('-')
    #   sueldo = int(sueldo)
    #   return cls(nombre, apellido, sueldo)
    
    def __str__(self):
        '''
        Si se ejecuta 'print(instancia de la clase)' ó 'str(instancia)' se correrá esta función. Sir
        ve para dar información respecto a lo que ya corrio la instancia.
        '''
        texto  = f'Ajuste lineal:\n -Parámetros de ajuste: {self.parametros}.\n'
        texto += f' -Matriz de covarianza de los parámetros: {self.cov_parametros}.\n'
        self.bondad()
        return texto

    def fit(self, modelo:str, **kwargs):
        '''
        Actualiza los coeficientes del ajuste y su matriz de covarianza acorde al modelo especificad
        o.

        INPUT:
        modelo: str: se deberá elegir entre 'regresion_lineal' ó 'curve_fit' de la librería scipy. S
        i se elige la última, se deberá incluir, además, el siguiente parámetro.
        
        expr: str: función de ajuste expresada como str. Es necesario que los parámetros tomen la fo
        rma 'letra_i' con i = 0, 1, 2, 3... Se debe respetar el orden en las varibles y para funcion
        es especiales usar la libreria numpy. Ejemplo:
        >> expr = 'a_0 + a_1*x**2 + np.sin(a_2*x)**3'

        **kwargs
        REGRESIÓN LINEAL:
        n: int: orden de la regresión.
        ordenada: bool: si incluir o no la ordenada al origen como parámetro a determinar. Si ordena
        da=False entonces se asume que es nula.

        CURVE FIT:https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.curve_fit.html
        p0: np.array: valores iniciales de los parámetros. Deben incluirse todos, de no pasar este p
        arámetro se tomara np.ones(shape = self.parametros.shape).
        
        bounds: 2-tuple or array_like: una tupla en la que en el primer valor se incluye un np.array 
        con la misma dimensión que los parámetros en la que irán las cotas inferiores; en el segundo
        valor, las superiores. Para que no haya cota usar, respectivamente, -np.inf y np.inf.
        
        method: {'lm', 'trf', 'dogbox'}: método usado para optimizar. Default es lm para problemas s
        in restricciones y trf si se proveen bounds. lm no funcionará si el numero de observaciones 
        es menor que el número de variables. En ese caso usar trf o dogbox.
        
        absolute_sigma: bool: si es True, la desviación estándar es usada en sentido absoluto y la m
        atriz de covarianza de los parámetros (self.cov_parametros) refleja valores absolutos. Si es
        False (default) sólo valores relativos de la desviación importan, en ese caso la matriz de c
        ovarianza se basa en un escaleo por un factor constante. 
        This constant is set by demanding that the reduced chisq for the optimal parameters popt whe
        n using the scaled sigma equals unity. In other words, sigma is scaled to match the sample v
        ariance of the residuals after the fit. Default is False. Mathematically, pcov(absolute_sigm
        a=False) = pcov(absolute_sigma=True) * chisq(popt)/(M-N).
        '''        
        # Chequea que el modelo especificado exista
        if modelo in Ajuste.modelos:
            pass
        else:
            raise Exception(f"No existe el modelo {modelo}")

        if modelo == 'regresion_lineal':
            self.parametros, self.cov_parametros = self.regresion_lineal(**kwargs)

            # Los datos predecidos por el modelo
            self.y_modelo = np.dot(self.vander, self.parametros)

        elif modelo == 'curve_fit':
            self.expr = kwargs['expr']

            # Define una función en base a la expresión especificada en los parámetros
            func = Ajuste.define_func(self.expr)

            # Se elimina la variable para prevenir que la lea curve_fit (no es un parámetro)
            kwargs.pop('expr')
            
            if self.x.shape[1]==1:
                self.parametros, self.cov_parametros = curve_fit(
                f = func,
                xdata = self.x.reshape(-1),
                ydata = self.y.reshape(-1),
                sigma = self.cov_y, # se puede pasar sigma en lugar de cov_y
                **kwargs) 
            else:
                self.parametros, self.cov_parametros = curve_fit(
                f = func,
                xdata = self.x,
                ydata = self.y,
                sigma = self.cov_y, # se puede pasar sigma en lugar de cov_y
                **kwargs) 
            
            # Los datos predecidos por el modelo
            self.y_modelo = func(self.x, *self.parametros)
    
    def regresion_lineal(self, n:int = 1, ordenada:bool = False):
        '''
        Realiza una regresión lineal y actualiza la matriz de Vandermonde.

        INPUT:
        n: int: orden de la regresión.
        
        ordenada: bool: si incluir o no la ordenada al origen como parámetro a determinar. Si ordena
        da=False entonces se asume que es nula.

        OUTPUT:
        parametros: np.array: los parámetros del ajuste.
        
        cov_parametros: np.array: la matriz de covarianza de los parámetros del ajuste.
        '''        
        # Matriz de Vandermonde:
        pfeats = PolynomialFeatures(degree = n, include_bias = ordenada)
        vander = pfeats.fit_transform(self.x)
        self.vander = vander.copy()
        
        # Calculos auxilares:
        inversa_cov = np.linalg.inv(self.cov_y)
        auxiliar = np.linalg.inv(np.dot(np.dot(vander.T, inversa_cov), vander))

        # Parámetros [At.Cov-1. A]-1.At.Cov-1.y = [a_0, a_1, ..., a_n]t
        parametros = np.dot(np.dot(np.dot(auxiliar, vander.T), inversa_cov), self.y) 

        # Matriz de covarianza de los parámetros [At.Cov-1.A]-1
        cov_parametros = np.linalg.inv(np.dot(vander.T, np.dot(inversa_cov, vander)))
           
        return parametros, cov_parametros
    
    def bondad(self):
        '''
        Calcula coeficientes para determinar la bondad del ajuste: r, R^2, Xi^2 y Xi^2 reducido.

        - La correlación lineal (r) me habla de qué tan determinadas están unas variables respecto a
        las otras. r(x,y) := cov(x,y)/sigma(x)*sigma(y) con -1 < r < 1. Si r = 1 las variables están
        correlacionadas (i.e: cuando una crece, la otra también), si r = -1 las variables están anti
        correlacionadas (i.e: cuando una crece, la otra decrece); si r = 0 no están correlacionadas.

        Lo que devuelve este método es la matriz de correlación entre las variables x e y. Es decir,
        r(x,x), r(x,y) = r(y,x) y r(y,y). Obviamente, el dato interesante es el que está fuera de la
        diagonal (es el que se reporta).

        - El coeficiente de determinación (R^2) determina la varianza en los resultados que puede ex
        plicarse con el modelo propuesto. Básicamente da una noción de si el ajuste cae o no dentro 
        del error de los datos. En el caso de un ajuste lineal, el r(x,y) coincide con el R^2.
        
        - El estadístico Chi cuadrado (Xi^2) sirve para testear que los datos siguen una distribución
        gaussiana cuya media en los datos adquiridos. La idea es contrastarlo contra su esperanza. Un 
        valor alto podría indicar errores subestimados o que los datos no siguen la distribución espe
        cificada. Mientras que un valor bajo podría indicar que los errores están sobre estimados. 
        ---> E(Xi^2) = (# de datos - # parametros ajustados) 
        ---> Sigma(Xi^2) = sqrt(2(# de datos - # parametros ajustados)).
        
        - La idea con el Chi cuadrado reducido (Xi^2 reducido) es la misma, pero su valor es más acot
        ado, puesto que por definición va normalizado por la esperanza de la Xi^2.
        ---> E(Xi_r^2) ~ 1
        ---> Sigma(Xi_r^2) = sqrt(2)/(# de datos - # parametros ajustados)**(3/2).
        '''    
        # Si el dominio es multidimensional no sé que parámetros determinan la bondad
        if self.x.shape[1] == 2:
            return print('Dominio multidimensional: los parámetros de bondad utilizados no sirven. FALTA IMPLEMENTAR')
            
        # Matriz de correlación lineal de los datos
        self.r = np.corrcoef(self.x.flatten(), self.y.flatten())

        # Coeficiente de determinación 1 - sigma_r**2/sigma_y**2
        sigma_r = self.y - self.y_modelo
        ss_res = np.sum(a = sigma_r**2)
        ss_tot = np.sum(a = (self.y - np.mean(self.y))**2) # es como el sigma_y pero experimental
        self.R2 = 1 - (ss_res / ss_tot)

        # Chi^2
        self.chi_2 = np.sum(((self.y - self.y_modelo)/self.sigma_y)**2, axis = 0)
        expected_chi_2 = len(self.y) - len(self.parametros)
        variance_chi_2 = np.sqrt(2*expected_chi_2)
        self.reduced_chi_2 = self.chi_2/(len(self.y)-len(self.parametros))

        # Printea la bondad
        texto  = f'Bondad del ajuste:\n -Correlación lineal: {self.r}.\n'
        texto += f' -R cuadrado: {self.R2}.\n'
        texto += f' -Chi cuadrado: {self.chi_2}. Expected {expected_chi_2} ± {variance_chi_2}'
        texto += f' -Chi cuadrado reducido: {self.reduced_chi_2}.'
        print(texto)

    def graph(self, estilo:str, label_x: str = 'x', label_y: str = 'y', alpha: float = 1000, save:bool = False, path:str = os.path.join(os.getcwd() + os.path.normpath('/ajuste.png'))):
        '''
        Grafica los datos ajustados de distintas formas.
        -'errores': realiza un gráfico con los residuos de cada dato.
        -'ajuste_1': realiza un gráfico del tipo errorbar con los datos y la curva del ajuste en
        cima.
        -'ajuste_2': realiza un gráfico del tipo errorbar con los datos y la curva del ajuste en
        cima. Además, se incluye el error del ajuste estimado asumiendo que en cada punto los da
        tos siguen la distribución propuesta.
        -'slider': realiza un gráfico interactivo del ajuste en el cual se pueden variar los par
        ámetros para ver con qué valor es conveniente inicializar el ajuste.

        estilos_graficos = ['ajuste_1', 'ajuste_2', 'slider', 'errores'] SLIDER NO IMPL
                Realiza una regresión lineal y actualiza la matriz de Vandermonde.

        INPUT:
        estilo: str: qué gráfico hacer. Se puede seleccionar entre: 'ajuste_1', 'ajuste_2', 'errore
        s' y 'slider'.
        
        label_x: str: nombre del label del eje horizontal.
        
        label_y: str: nombre del label del eje horizontal.
        
        alpha: float: es el factor multiplicativo con el que se genera la tira auxiliar con la cual
        se grafica la curva del ajuste. En general si la cantidad de puntos es la misma que la cant
        idad de datos. La linea es entrecortada y no se ve suave. Se define explícitamente como x = 
        np.linspace(self.x[0], self.x[-1], len(self.x)*alpha). Es decir, que si alpha = 1.0, entonc
        es la cantidad de puntos que tendrá la curva es la misma que la cantidad de datos.
        
        save: bool: si se guarda o no el gráfico. Por defecto se guardarán en el directorio en el c
        ual se está ejecutando el código con el nombre 'ajuste.png'. De querer, se puede especifica
        r el el nombre del archivo y su path con el siguiente parametro

        path: str: donde y con qué nombre guardar la imagen

        OUTPUT:
        parametros: np.array: los parámetros del ajuste.
        
        cov_parametros: np.array: la matriz de covarianza de los parámetros del ajuste.
        '''
        # Chequeo que el estilo este dentro de los disponibles
        if estilo in Ajuste.estilos_graficos:
            pass
        else:
            var = ', '.join(Ajuste.estilos_graficos)
            raise Exception(f"No existe el gráfico {estilo}. Los disponibles son: {var}")

        if estilo == 'errores':
            # Calculo los residuos
            residuos = self.y_modelo - self.y
            
            # Creo la figura y grafico
            fig, ax = plt.subplots(nrows = 1, ncols = 1)
            eje_x = np.zeros(shape = self.y.shape)
            ax.scatter(x = self.x, y = residuos, s = 10, color = 'black', label = 'Resiudos')
            ax.plot(self.x, eje_x, color = 'black', linewidth = 0.4)
            ax.vlines(x = self.x, ymin = eje_x, ymax = eje_x + residuos, color = 'red', alpha = .8)
            ax.set_xlabel(xlabel = label_x)
            ax.set_ylabel(ylabel = label_y)
            ax.grid()
            ax.legend()
            fig.tight_layout()
            if save:
                fig.savefig(path)
            else:
                fig.show()

        elif estilo == 'ajuste_1':
            
            # Ajuste si se usa curve_fit o regresion_lineal
            if self.expr is not None:
                # Tiras auxiliares para graficar, tomo más puntos que los datos
                x_auxiliar = np.linspace(self.x[0], self.x[-1], len(self.x)*alpha).reshape(-1)
                ajuste = Ajuste.define_func(self.expr)(x_auxiliar, *self.parametros).reshape(-1)
            else:
                # Tiras auxiliares para graficar, tomo más puntos que los datos
                x_auxiliar = np.linspace(self.x[0], self.x[-1], len(self.x)*alpha).reshape(-1)
                if len(self.parametros) == 1:
                    ajuste = self.parametros[0]*x_auxiliar
                else:
                    # Si se pasa ordenada = True
                    ajuste = self.parametros[0] + self.parametros[1]*x_auxiliar
                

            # Creo la figura y grafico
            fig, ax = plt.subplots(nrows = 1, ncols = 1)
            ax.scatter(x = self.x, y = self.y, s = 5, color = 'black', label = 'Datos')
            ax.errorbar(self.x.reshape(-1), self.y.reshape(-1), yerr = self.sigma_y.reshape(-1), marker = '.', fmt = 'None', capsize = 1.5, color = 'black', label = 'Error de los datos')
            ax.plot(x_auxiliar, ajuste, 'r-', label = 'Ajuste', alpha = .5)
            ax.set_xlabel(xlabel = label_x)
            ax.set_ylabel(ylabel = label_y)
            ax.grid()
            ax.legend()
            fig.tight_layout()
            if save:
                fig.savefig(path)
            else:
                fig.show()


        elif estilo == 'ajuste_2':
            # Ajuste si se usa curve_fit o regresion_lineal
            if self.expr is not None:
                # Tiras auxiliares para graficar, tomo más puntos que los datos
                x_auxiliar = np.linspace(self.x[0], self.x[-1], len(self.x)*alpha).reshape(-1)
                ajuste = Ajuste.define_func(self.expr)(x_auxiliar, *self.parametros).reshape(-1)
                variables_nom = np.unique([var.group() for var in re.finditer(pattern = '[a-z]\_\d', string = self.expr)]).tolist()
                variables = [(variables_nom[i], self.parametros[i]) for i in range(len(variables_nom))]
                
                # Calculo el error del ajuste para cada dato
                franja_error = Propagacion_errores(
                variables = variables, 
                errores = self.cov_parametros, 
                formula = self.expr, 
                dominio = x_auxiliar
                ).fit()[1]                
            else:
                # Tiras auxiliares para graficar, tomo más puntos que los datos
                x_auxiliar = np.linspace(self.x[0], self.x[-1], len(self.x)*alpha).reshape(-1)
                if len(self.parametros) == 1:
                    # El ajuste y las variables
                    ajuste = self.parametros[0]*x_auxiliar
                    variables = [('a_0',self.parametros[0])]

                    # Calculo el error del ajuste para cada dato
                    franja_error = Propagacion_errores(
                    variables = variables, 
                    errores = self.cov_parametros, 
                    formula = 'a_0*x_', 
                    dominio = x_auxiliar
                    ).fit()[1]                    
                else:
                    # Si se pasa ordenada = True
                    ajuste = self.parametros[0] + self.parametros[1]*x_auxiliar
                    variables = [('a_0',self.parametros[0]), ('a_1', self.parametros[1])]
            
                    # Calculo el error del ajuste para cada dato
                    franja_error = Propagacion_errores(
                    variables = variables, 
                    errores = self.cov_parametros, 
                    formula = 'a_0 + a_1*x_', 
                    dominio = x_auxiliar
                    ).fit()[1]
            
            # Creo la figura y grafico
            fig, ax = plt.subplots(nrows = 1, ncols = 1)
            ax.scatter(x = self.x, y = self.y, s = 5, color = 'black', label = 'Datos')
            ax.errorbar(self.x.reshape(-1), self.y.reshape(-1), yerr = self.sigma_y.reshape(-1), marker = '.', fmt = 'None', capsize = 1.5, color = 'black', label = 'Error de los datos')
            ax.plot(x_auxiliar, ajuste, 'r-', label = 'Ajuste', alpha = .5)
            ax.plot(x_auxiliar, ajuste + franja_error, '--', color = 'green', label = 'Error del ajuste')
            ax.plot(x_auxiliar, ajuste - franja_error, '--', color = 'green')
            ax.fill_between(x_auxiliar, ajuste - franja_error, ajuste + franja_error, facecolor = "gray", alpha = 0.3)
            ax.set_xlabel(xlabel = label_x)
            ax.set_ylabel(ylabel = label_y)
            ax.grid()
            ax.legend()
            fig.tight_layout()
            if save:
                fig.savefig(path)
            else:
                fig.show()
        elif estilo == 'slider':
            raise Exception('No implementado todavía')

    @staticmethod
    def define_func(expr:str):
        '''
        Cuenta el numero de variables de la forma 'a_i' y defune una función con 'x' como 
        variable independiente y los 'a_i' como coeficientes.
        
        INPUT:
        expr: str: formula de la función a fittear.
        '''        
        n_vars = ['x_'] + np.unique([var.group() for var in re.finditer(pattern = '[a-z]\_\d', string = expr)]).tolist()
        args = ', '.join(n_vars)
        expresion = expr
        exec(f'def func({args}):\n    import numpy as np\n    return {expresion}', locals())
        return locals()['func']

if __name__ == '__main__':

    # EJEMPLO 1D:
    x = np.linspace(0,100,40)
    # y = 1 + 2*np.sin(0.1*x)*np.exp(-.5*x)
    # y = x*0.25 + 3
    y = np.piecewise(x = x, condlist = [x<=40 , x>40], funclist = [lambda x : 2*np.exp(-.04*x),.4])
    sigma = .05
    signal = np.random.normal(y, sigma, size = y.shape)
 
    # expr = 'a_0 + a_1*np.sin(.1*x_)*np.exp(a_2*x_)'
    expr = 'np.piecewise(x = x_, condlist = [x_<=a_0 , x_>a_0], funclist = [lambda x_ : a_1*np.exp(-a_2*x_),a_3])'

    aj = Ajuste(x, signal, cov_y = np.array([sigma for i in x]).reshape(-1,1))
    # aj.fit(modelo='curve_fit', expr = expr, bounds =([-10, 0, -np.inf] , [10, 10, 0]))
    aj.fit(modelo='curve_fit', expr = expr, p0 = [40,2,.04,.4], method = 'dogbox')
    # aj.fit(modelo = 'regresion_lineal', ordenada = True)
    aj.graph(estilo = 'ajuste_1')
    aj.graph(estilo = 'errores', label_x = 'Tiempo [s]', label_y = r'Tension [$\propto V$]')
    aj.graph(estilo = 'ajuste_2', label_x = 'Tiempo [s]', label_y = r'Tension [$\propto V$]')
    aj.parametros
    aj.bondad()
    aj.cov_parametros

    # # EJEMPLO CON DOMINIO 2D
    # x_1 = np.linspace(0, 100, 70).reshape(-1,1)
    # x_2 = np.linspace(0, 54, 70).reshape(-1,1)
    # X = np.concatenate([x_1, x_2], axis = 1)
    # pfeats = PolynomialFeatures(degree = 1, include_bias =True)
    # vander = pfeats.fit_transform(X)
    # y, sigma = (10 + X[:, 0]*2+ X[:, 1]*9).reshape(-1,1), .01
    # signal = np.random.normal(y, sigma, size = y.shape)


    # def func(x_:np.array, a:np.float64, b:np.float64, c:np.float64):
    #     return a*x_[:, 0] + b* x_[:, 1] + c
    # regr = Ajuste(x = X, y = signal)   
    # regr.fit(modelo = 'curve_fit', expr = 'a_0*x_[:, 0] + a_1* x_[:, 1] + a_2')
    # # regr.fit(modelo = 'regresion_lineal', ordenada = True)
    # regr.parametros
    # final, dato = -1, 1
    # plt.scatter(X[:final,dato], signal[:final])
    # plt.plot(X[:final,dato], regr.y_modelo[:final], color = 'red')
    # plt.show(block = False)    
    


import numpy as np
from sympy import symbols, lambdify, diff
from sympy import log, sin, exp, cos, tan, sqrt, pi, atan, asin, acos, Abs, cot, sec, csc, sinc, acot, asec, acsc, atan2, sinh, cosh, tanh, coth, sech, csch, asinh, acosh, atanh, acoth, asech, root, Piecewise

class Propagacion_errores:

    def __init__(self, formula:str, variables: list, errores: np.array, dominio: np.array = None) -> None:
        '''
        INPUT:
        formula: str: la fórmula de la variable cuyo error se quiere propagar. Esta debe usar exclu
        sivamente la librería de numpy o sympy para funciones especiales (np.sin, cos, etc.).
        
        variables: list: es una lista de tuplas de dos elementos. Cada tupla esta compuesta por dos
        valores. El primer valor es un str cuyo símbolo representa a cada variable (se debe respeta
        r las mismas variables usadas en la fórmula).
        
        errores: np.array: array con los errores de cada variable o matriz de covarianza de las var
        iables. Se debe respetar el orden pasado en 'variables' y 'formula'. Las dimensiones del ar
        ray deben ser (cantidad_de_var, 1) o (cantidad_de_var, cantidad_de_var) si se pasa la matri
        z de covarianza.
        '''
        # Variables importantes e hiperparámetros usados dentro de la clase
        self.formula = formula
        self.variables = variables
        self.dominio = dominio
        self.valor = None
        self.error = None
        
        # Determina la matriz de covarianza:
        if errores.ndim != 2:
            raise ValueError("El array que se paso no tiene las dimensiones requeridas.\n Debería tener dimensiones (len(variables), len(variables)) ó (len(variables), 1) si se pasan desviaciones estándar.")

        if errores.shape[1] == 1:
            self.covarianza = np.diag(errores.reshape(-1)**2)
            self.errores = errores
        else:
            self.covarianza = errores
            self.errores = np.sqrt(np.diag(errores))

    def __str__(self):
        '''
        Si se ejecuta 'print(instancia de la clase)' ó 'str(instancia)' se correrá esta función. Sir
        ve para dar información respecto a lo que ya corrio la instancia.
        '''
        texto  = f'El valor obtenido es: ({self.valor} ± {self.error})'
        print(texto)
        return texto
    
    # @classmethod 
    # def without_variables(cls, dominio: np.array, formula:str, variables: list, errores: np.array)):
    #     '''
    #     Esta forma de llamar a propagación está implementada para evaluar el error de una función en
    #     una tira de datos. Por ejemplo: f(x_) = a_0*x_**2. En este caso la variable x_ está fija, pe
    #     ro corre a lo largo de muchos puntos. Mientras que a_0 es un parametro con error.
    #     # FIXME TODO: documentar bien
    #     En este tipo de llamado se reserva la variable 'x_' para ser evaluada en 'dominio'.
    #     '''
        
    #     return cls(variables = variables, errores = errores, formula = formula)

    def fit(self):
        '''
        Calcula el error propagado y el valor de la magnitud de interés.
        
        OUTPUT: tupla con dos valores.
        self.valor: np.float64: valor de la cantidad pasada en la fórmula.
        
        self.error: np.float64: error (raíz cuadrada de varianza) propagado de la cantidad pasada e
        n la fórmula.
        '''
        simbolos = [i for i,j in self.variables]
        valores = [j for i,j in self.variables]
        
        # Defino como símbolos las variables de las cuales depende la expresión:
        for sim in simbolos:
            globals()[sim] = symbols(sim, real = True)

        # Si se pasa dominio se agrega la variable x_ que es muda
        if isinstance(self.dominio, np.ndarray):
            x_ = symbols('x_', real = True)
        
        # Defino la expresión simbólica:
        self.formula = Propagacion_errores.traductor_numpy_sympy(self.formula)
        formula = eval(self.formula)
        
        # Calculo la aproximación lineal de la covarianza de el i-ésimo dato con el j-ésimo y sumo:
        derivadas_parciales = [diff(formula, eval(sim)) for sim in simbolos]
        covarianza_resultado = 0
        for i in range(len(simbolos)):
            for j in range(len(simbolos)):
                covarianza_resultado += derivadas_parciales[i]*self.covarianza[i, j]*derivadas_parciales[j]
        
        # Fórmula del error simbólico
        error_simbolico = sqrt(covarianza_resultado)

        # Convierto la expresión simbólica en un módulo numérico (numpy) para poder reemplazar:
        if isinstance(self.dominio, np.ndarray):
            lambd_err = lambdify(['x_'] + simbolos, error_simbolico, modules = ['numpy'])
            lambd_val = lambdify(['x_'] + simbolos, formula, modules = ['numpy'])
            valores = [self.dominio] + valores
            self.valor, self.error = lambd_val(*valores), lambd_err(*valores)

            # Borro las variables auxiliares que se definieron en globals()
            for sim in simbolos:
                del globals()[sim]

            return self.valor, self.error
        
        else:
            lambd_err = lambdify(simbolos, error_simbolico, modules = ['numpy'])
            lambd_val = lambdify(simbolos, formula, modules = ['numpy'])
            self.valor, self.error = lambd_val(*valores), lambd_err(*valores)

            # Borro las variables auxiliares que se definieron en globals()
            for sim in simbolos:
                del globals()[sim]

            return self.valor, self.error

    @staticmethod
    def traductor_numpy_sympy(expr:str):
        '''
        Traductor de fórmulas escritas en strings de numpy a sympy.

        INPUT:
        expr: str: expresión a traducir desde el paquete numpy al sympy.
        
        OUTPUT: 
        expr: str: misma expresión, pero escrita en formato sympy
        '''
        # Por si toma una función partida 
        if 'np.piecewise' in expr:
            
            # Extraigo rangos
            start = expr.find('[') + 1
            end = expr.find(']')
            list_of_ranges = expr[start:end].replace(' ', '').split(',')
            
            # Extraigo funciones
            start = expr.rfind('[') + 1
            end = expr.rfind(']')
            list_of_funcs_aux = expr[start:end].replace(' ', '').split(',')
            list_of_funcs = []
            for fun in list_of_funcs_aux:
                if 'lambda' in fun:
                    list_of_funcs.append(fun.split(':')[1])
                else: 
                    list_of_funcs.append(fun)
            
            # Creo los argumentos de la funcion de Sympy
            args = ''
            for range, func  in zip(list_of_ranges, list_of_funcs):
                args+=f'({func}, {range}),'
            expr = f'Piecewise({args})'
        
        expr = expr.replace('np.', '')
        dic = {
            'arctan':'atan',
            'arctan2':'atan2',
            'arcsin':'asin',
            'arccos':'acos',
            'abs': 'Abs',
            'arcsinh': 'asinh',
            'arccosh': 'acosh',
            'arctanh': 'atanh',
        }
        for key in dic.keys():
            if key in expr:
                expr = expr.replace(key, dic[key])
        return expr

# Extraigo rangos
if __name__ == '__main__':
    # # Prueba regresión lineal
    # # INPUT
    # expr = 'A*cos(f*t) + C'    
    # variables = [
    #     ('f', 100), # Hz
    #     ('A', 2), # Volts
    #     ('t', 1), # s
    #     ('C', .5), # Volts
    #     ]
    # errores = np.array([0.0005, .0001, 0, .0001]).reshape(-1,1) # Notar que al tiempo no le asigne error
    # propaga = Propagacion_errores(formula = expr, variables = variables, errores = errores)
    # valor, error = propaga.fit()
    # print(valor, error)

    remanencia = [0.6617640000000007, 0.648, 0.6463663999999998, 0.644, 0.6446468571428572, 0.6313968000000003, 0.6101840000000002, 0.6201240000000001, 0.5880179999999999, 0.5917687999999999, 0.5733999999999997, 0.5831424000000001, 0.5679799999999995, 0.5564127999999998, 0.5430304000000001, 0.5362119999999997, 0.5105864, 0.517164, 0.4968688, 0.48230079999999953, 0.4683824, 0.46571199999999935, 0.4647747999999996, 0.4464924000000006, 0.42020479999999993, 0.4440120000000002, 0.4033028000000004, 0.4176367999999996, 0.38584560000000045, 0.3707120000000005, 0.38518199999999947, 0.36424320000000016, 0.33886479999999963, 0.3343535999999997, 0.3384912000000001, 0.3169108571428574, 0.2907544000000004, 0.2953548000000006, 0.2602840000000004, 0.2610116, 0.20596000000000017, 0.22832840000000082, 0.21134320000000062, 0.19457520000000011, 0.18572639999999896, 0.18100000000000027, 0.11647039999999967, 0.12683519999999993, 0.11390000000000014, 0.1225975999999997, 0.0967327999999996, 0.09915520000000044, 0.08739759999999902, 0.08858000000000059, 0.06923520000000026, 0.04971599999999991, 0.04452320000000015, 0.04711520000000016, 0.03800000000000004, 0.016878399999999738, 0.006257599999999929, 0.019350400000000274, 0.014008000000000338, 0.011264999999999938, 0.017362799999999897, 0.010928399999999842, 0.011608000000000238, 0.013312000000000067, 0.0080819999999997, 0.013582666666666762, 0.012048333333332904, 0.00451599999999988, 0.008423619047618899, 0.009185523809523857, 0.00942560000000025, 0.00749600000000034, 0.009920400000000086, 0.010071333333333219, 0.0107526666666668, 0.0051593333333333205, 0.009077600000000165, 0.008663111111111118, 0.008011047619047704, 0.011247999999999411, 0.009370000000000028, 0.00923333333333337, 0.010202000000000128, 0.008650399999999874, 0.007635600000000099, 0.008488833333333144, 0.007483333333333452, 0.010556571428571316, 0.008483142857142578, 0.008123809523809191, 0.008267200000000186, 0.008483200000000425, 0.007842400000000128, 0.0086675000000002, 0.004449600000000178, 0.005907000000000015]
    
    # Estoy haciendo la resta entre dos valores del canal 2, entonces aparece el factor sqrt(2):
    error = np.full(len(remanencia), 2.77680184e-08)

    # Hago el ajuste
    formula = 'np.piecewise(x_, [x_ < a_0, x_ >= a_0], [lambda x_: a_1*np.abs(x_- a_0)**(a_2) + a_3, a_3])'
    formula = formula.replace('x_', str(float(3)))
    variables = [('a_0',2.48909942e+02), ('a_1',6.70276581e-02),('a_2',4.48016420e-01),('a_3',1.42274642e-02)]
    print(Propagacion_errores(formula = formula, variables = variables, errores = error.reshape(-1,1)).fit())


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['labos']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.2.2',
 'numpy>=1.21',
 'pandas>1.4',
 'requests>=2.2',
 'scikit-learn>=1.1',
 'scipy>=1.10.0',
 'sympy>=1.10']

setup_kwargs = {
    'name': 'labos',
    'version': '0.1.2.5',
    'description': 'Paquete para trabajar en el labo',
    'long_description': '# Labos\n\nEste paquete está dedicado principalmente a facilitar el uso de herramientas y paquetes utilizados en la adquisición y procesamiento de datos en el marco de los laboratorios de la carrera en [*Ciencias Físicas de la UBA*](https://www.df.uba.ar/es/).\n\nLos paquetes hasta ahora implementados son de adquisición y análisis de datos.\n\n<!-- headings -->\n## Análisis de datos\n\n### Propagación\nEste paquete está dedicado a propagar errores utilizando una aproximación lineal sobre la formula de covarianza. El paquete está basado en las librerías numpy e sympy.\n```python\nfrom labos.propagacion import Propagacion_errores\nimport numpy as np\n\n# Formula a propagar\nexpr = \'A*cos(f*t) + C\'    \n\n# Las variables dependientes y sus valores\nvariables = [\n    (\'f\', 100), # Hz\n    (\'A\', 2), # Volts\n    (\'t\', 1), # s\n    (\'C\', .5), # Volts\n    ]\n\n# Los errores de las variables\nerrores = np.array(\n    [.0005,\n    .0001, \n    0, \n    .0001]\n    ).reshape(-1,1)\n\n# Instancia de la clase\npropaga = Propagacion_errores(\n    formula = expr,\n    variables = variables,\n    errores = errores)\npropaga.fit()\n>>(2.224637744575368, 0.0005232992460070357)\n```\n### Ajuste\nEste paquete está dedicado a realizar ajustes sobre datos. El paquete está basado en las librerías numpy e sympy.\n\n#### Heading 3\n##### Heading 4\n###### Heading 5\n\n<!-- line breaks -->\n<!-- ENTER -->\n\n<!-- Italics -->\nThis is an *italic* text\n\nThis is an _italic_ text\n\n<!-- Strongs -->\nThis is an **strong** text\n\nThis is an __strong__ text\n\n<!-- StrikeTrough -->\nThis is an ~~striketrogh~~ text\n\n\n\n<!-- UL -->\n* item 1\n* item 2\n* item 3\n    * item 3.1\n    * item 3.2    \n* item 4\n\n<!-- OL -->\n1. item 1\n1. item 2\n1. item 3\n1. item 4\n\n<!-- Links -->\n[Faztweb.com](https://www.faztweb.com)\n\n[Faztweb.com](https://www.faztweb.com "Custom title")\n<!-- Blockquote -->\n> this is a quote\n\n<!-- Horizontal Rule -->\n___\n---\n\n<!-- Inline code -->\n`console.log(\'hello world\')`\n\n`<h1>Hello world</h1>`\n\n<!-- IMAGES -->\n<!-- ![Vscode Logo](https://upload.wikimedia.org/wikipedia/commons/thumb/9/9a/Visual_Studio_Code_1.35_icon.svg/1200px-Visual_Studio_Code_1.35_icon.svg.png) -->\n\n<!-- ![Vscode logo](./vscode.png "vscode") -->\n\n<!-- GITHUB MD -->\n\n```\nfirst line of code\nsecond line of code\n```\n\n```python\nprint("hello world")\n```\n\n```javascript\nconsole.log(\'hello world\')\n\nconst test = (str) => str + \'test\';\n```\n\n```html\n<h1>Hello World</h1>\n```\n\n<!-- TABLES -->\n| Product       | Price         |quantity   |\n| ------------- |:-------------:| :--------:|\n| Laptop        | 3.33          | 2         |\n| Mouse         | 10.33         | 1         |\n\n* [x] task1\n* [] task2\n* [] task3\n* [x] task4\n\n<!-- Mentiosn -->\n@faztweb :+1: :smile:\n',
    'author': 'joctavio287',
    'author_email': 'joctavio287@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)
